/* sane - Scanner Access Now Easy.
   Copyright (C) 1997 David Mosberger-Tang/Peter Kirchgessner
   This file is just for testing HP-Scanners using SCL.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file implements a program that allows to send various SCL-commands
   to HP-scanner and start a scan.

   V1.10, PK, 01-Sep-2001 Add -data <datafile>
   V1.11, PK, 25-Aug-2002 Add buttonstate
   V1.12, PK, 06-Jan-2006 Handle grayscale scanning

   *NO WARRANTY*   */

#include <sane/config.h>

#include <limits.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include <sane/sane.h>
#include <sane/saneopts.h>
#include <sane/sanei_scsi.h>

#define BACKEND_NAME	hpscan
#include <sane/sanei_backend.h>

#ifndef PATH_MAX
# define PATH_MAX	1024
#endif

#include <sane/sanei_config.h>

#include "hptools.h"

#define HP_BUFSIZE (128*1024)

#define INQ_LEN	255
static const u_int8_t inquiry[] =
  {
    0x12, 0x00, 0x00, 0x00, INQ_LEN, 0x00
  };

static const u_int8_t test_unit_ready[] =
  {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00
  };

static char *add_str = "";
static char output[1024] = { "out.pnm" };

static struct INQCMD {
  char *key;
  char *dsc;
  char *code;
  int inq_id;
}
inq_cmd[] =
{
    { NULL,     NULL,                     "\033*a%dA", 10306, },
    { "abkgrd", "Automatic Bkgrd",        "\033*a%dB", 10307, },
    { "compr",  "Compression",            "\033*a%dC", 10308, },
    { "dtype",  "Download Type",          "\033*a%dD", 10309, },
    { "xscale", "X Scale",                "\033*a%dE", 10310, },
    { "yscale", "Y Scale",                "\033*a%dF", 10311, },
    { "dwidth","Select Data Width",       "\033*a%dG", 10312, },
    { NULL,     NULL,                     "\033*a%dH", 10313, },
    { "invers", "Inverse Image",          "\033*a%dI", 10314, },
    { "bwdpat", "B/W Dither Pattern",     "\033*a%dJ", 10315, },
    { "contr",  "Contrast Level",         "\033*a%dK", 10316, },
    { "intens", "Intensity Level",        "\033*a%dL", 10317, },
    { "mirror", "Mirror Image",           "\033*a%dM", 10318, },
    { NULL,     NULL,                     "\033*a%dN", 10319, },
    { "res1",   "Reserved1",              "\033*a%dO", 10320, }, /* scl.c */
    { "xextd",  "Window X Ext (Dec)",     "\033*a%dP", 10321, },
    { "yextd",  "Window Y Ext (Dec)",     "\033*a%dQ", 10322, },
    { "xres",   "X Res",                  "\033*a%dR", 10323, },
    { "yres",   "Y Res",                  "\033*a%dS", 10324, },
    { "otyp",   "Output Data Type",       "\033*a%dT", 10325, },
    { NULL,     NULL,                     "\033*a%dU", 10326, },
    { "speed",  "Scan speed",             "\033*a%dV", 10327, }, /* scl.c */
    { NULL,     "Download Binary",        "\033*a%dW[]", 10328},
    { "xposd",  "Window X Pos (Dec)",     "\033*a%dX", 10329, },
    { "yposd",  "Window Y Pos (Dec)",     "\033*a%dY", 10330, },
    { "rcon",   "Reconnect level",        "\033*a%dZ", 10331, }, /* scl.c */
    { "nviews", "Select Number of Views", "\033*f%dA", 10466, },
    { "view",   "Select View",            "\033*f%dB", 10467, },
    { "preladf","Preload ADF",            "\033*f%dC", 10468, },
    { "media",  "Media type",             "\033*f%dD", 10469, }, /* scl.c */
    { "csys",   "Coordinate system",      "\033*f%dE", 10470, }, /* scl.c */
    { "pos2",   "Pos Secondary Elem",     "\033*f%dF", 10471, },
    { NULL,     NULL,                     "\033*f%dG", 10472, },
    { NULL,     NULL,                     "\033*f%dH", 10473, },
    { NULL,     NULL,                     "\033*f%dI", 10474, },
    { NULL,     NULL,                     "\033*f%dJ", 10475, },
    { NULL,     NULL,                     "\033*f%dK", 10476, },
    { "ltest",  "Light Source Test",      "\033*f%dL", 10477, },
    { NULL,     NULL,                     "\033*f%dM", 10478, },
    { NULL,     NULL,                     "\033*f%dN", 10479, },
    { NULL,     NULL,                     "\033*f%dO", 10480, },
    { "xext",   "Window X Ext (Pix)",     "\033*f%dP", 10481, },
    { "yext",   "Window Y Ext (Pix)",     "\033*f%dQ", 10482, },
    { NULL,     NULL,                     "\033*f%dR", 10483, },
    { "scan",   "Scan Window (w/ID)",     "\033*f0S", 10484, },
    { "test",   "Scanner Test",           "\033*f%dT", 10485, }, /* scl.c */
    { NULL,     NULL,                     "\033*f%dU", 10486, },
    { NULL,     NULL,                     "\033*f%dV", 10487, },
    { NULL,     NULL,                     "\033*f%dW", 10488, },
    { "xpos",   "Window X Pos (Pix)",     "\033*f%dX", 10489, },
    { "ypos",   "Window Y Pos (Pix)",     "\033*f%dY", 10490, },
    { NULL,     NULL,                     "\033*f%dZ", 10491, },
    { NULL,     NULL,                     "\033*s%dA", 10882, },
    { NULL,     NULL,                     "\033*s%dB", 10883, },
    { NULL,     NULL,                     "\033*s%dC", 10884, },
    { NULL,     NULL,                     "\033*s%dD", 10885, },
    { NULL,     NULL,                     "\033*s%dE", 10886, },
    { NULL,     NULL,                     "\033*s%dF", 10887, },
    { NULL,     NULL,                     "\033*s%dG", 10888, },
    { NULL,     NULL,                     "\033*s%dH", 10889, },
    { NULL,     NULL,                     "\033*s%dI", 10890, },
    { NULL,     NULL,                     "\033*s%dJ", 10891, },
    { NULL,     NULL,                     "\033*s%dK", 10892, },
    { NULL,     NULL,                     "\033*s%dL", 10893, },
    { NULL,     NULL,                     "\033*s%dM", 10894, },
    { NULL,     NULL,                     "\033*s%dN", 10895, },
    { NULL,     NULL,                     "\033*s%dO", 10896, },
    { NULL,     NULL,                     "\033*s%dP", 10897, },
    { NULL,     NULL,                     "\033*s%dQ", 10898, },
    { NULL,     NULL,                     "\033*s%dR", 10899, },
    { NULL,     NULL,                     "\033*s%dS", 10900, },
    { NULL,     NULL,                     "\033*s%dT", 10901, },
    { NULL,     "Upload Binary",          "\033*s%dU", 10902, },
    { NULL,     NULL,                     "\033*s%dV", 10903, },
    { NULL,     NULL,                     "\033*s%dW", 10904, },
    { NULL,     NULL,                     "\033*s%dX", 10905, },
    { NULL,     NULL,                     "\033*s%dY", 10906, },
    { NULL,     NULL,                     "\033*s%dZ", 10907, },
    { "calarea","Select Cal Area (Y-pos)","\033*u%dA", 10946, },
    { "order",  "Byte order",             "\033*u%dB", 10947, },
    { "calstrip","Select Cal Strip Parm.", "\033*u%dC", 10948, },
    { "xpascan","XPA Scan Window (w/ID)", "\033*u0D", 10949, },
    { "speedmode","Speed Mode",             "\033*u%dE", 10950, },
    { "filter", "Set Filter",             "\033*u%dF", 10951, },
    { "calibmode","Calibration mode",       "\033*u%dG", 10952, }, /* scl.c */
    { "xpa",    "XPA Disable/Enable",     "\033*u%dH", 10953, },
    { NULL,     NULL,                     "\033*u%dI", 10954, },
    { "coldpat","Color Dither Pattern",   "\033*u%dJ", 10955, },
    { "tonemap","Select Tone Map",        "\033*u%dK", 10956, },
    { "lpower", "Light Power",            "\033*u%dL", 10957, }, /* scl.c */
    { NULL,     NULL,                     "\033*u%dM", 10958, },
    { NULL,     NULL,                     "\033*u%dN", 10959, },
    { NULL,     NULL,                     "\033*u%dO", 10960, },
    { NULL,     NULL,                     "\033*u%dP", 10961, },
    { NULL,     NULL,                     "\033*u%dQ", 10962, },
    { "calib",  "Recalibrate (w/ID)",     "\033*u0R", 10963, },
    { "adfscan","ADF Scan Window (w/ID)", "\033*u0S", 10964, },
    { "mcoef",  "Select Matrix Coef",     "\033*u%dT", 10965, },
    { "unload", "Unload Document",        "\033*u0U", 10966, },
    { NULL,     NULL,                     "\033*u%dV", 10967, },
    { NULL,     NULL,                     "\033*u%dW", 10968, },
    { "change", "Change Document",        "\033*u0X", 10969, },
    { "adfbfeed","ADF Bkgrd Feed",        "\033*u0Y", 10970, },
    { NULL,     NULL,                     "\033*u%dZ", 10971, },
    { "clrerr", "Clear errors",           "\033*oE", -1, },
    { "reset", "Reset",                   "\033E", -1, }
};

static struct INQDEVPARAM {
    int inq_id;
    char *key;
    char *dsc;
}
inq_devparam[] =
{
    {    3, "hpmodel1", "HP Model #1" },
    {   10, "hpmodel2", "HP Model #2" },
    {    9, "hpmodel3", "HP Model #3" },
    {   11, "hpmodel4", "HP Model #4" },
    {   12, "hpmodel5", "HP Model #5" },
    {   14, "hpmodel6", "HP Model #6" },
    {   15, "hpmodel8", "HP Model #8" },
    {   16, "hpmodel9", "HP Model #9" },
    {   17, "hpmodel10","HP Model #10" },
    {   18, "hpmodel11","HP Model #11" },
    {   19, "hpmodel12","HP Model #12" },
    {   21, "hpmodel14","HP Model #14" },
    {   31, "hpmodel16","HP Model #16" },
    {   32, "hpmodel17","HP Model #17" },
    {    4, "firmware", "Firmware Date" },
    {    5, "scanstat", "Scan Status" },     /* scl.c */
    {    6, "adfstat",  "ADF Status" },      /* scl.c */
    /* {   21, "Reserved" },   */     /* scl.c */
    {   24, "adfcap",   "ADF Capability" },
    {   25, "adfbin",   "ADF Bin" },
    {   26, "open",     "Open Since last Chg" },
    {   27, "rdyunload","Ready to Unload" },
    {   28, "xpacap",   "XPA Capability" },
    {   29, "xpastat",  "XPA Status" },      /* scl.c */
    {   30, "xpardy",   "XPA Ready" },
    {  256, "errstack", "Max Error Stack" },
    {  257, "cuerrstack","Current Error Stack" },
    {  259, "cuerr",     "Current Error Number" },
    {  261, "olderr",    "Oldest Error" },
    { 1024, "pixelsl",   "Pixels/Scan Line" },
    { 1025, "bytesl",    "Bytes/Scan Line" },
    { 1026, "sl",        "Number Scan Lines" },
    { 1027, "adfrdy",    "ADF Ready" },
    { 1028, "devdpi",    "Device Pixels/Inch" },
    { 1029, "optres",    "Native Optical Resolution" },
    { 1031, "autofilt",  "Auto Filtering" },
    /* { 1034, "Should Be" },  */          /* scl.c */
    { 1037, "byteslast", "Bytes in last scan" },   /* scl.c */
    { 1040, "autospeed", "Inquire Auto Speed Mode" },
    { 1041, "whichadf",  "Which ADF" },            /* scl.c */
    { 1044, "buttonstate", "Inquire button state" },
    { 1045, "preloaded", "Inquire Preloaded" },
    { 1046, "autocoeff", "Auto coefficient" },     /* scl.c */
    { 1047, "scandir",   "Scan direction" }        /* scl.c */
};

static SANE_Status
read_data (int fd, u_char * dest, size_t *len)
{
  static u_char read_6[6];

  memset (read_6, 0, sizeof (read_6));
  read_6[0] = 0x08;
  read_6[2] = *len >> 16;
  read_6[3] = *len >> 8;
  read_6[4] = *len;
  return hptools_device_cmd (fd, read_6, sizeof (read_6), dest, len);
}

static SANE_Status
write_data (int fd, void *data, size_t len)
{
  char write_6[6+HP_BUFSIZE];

  memset (write_6, 0, 6);
  write_6[0] = 0xa;
  write_6[2] = len >> 16;
  write_6[3] = len >> 8;
  write_6[4] = len;
  memcpy (write_6 + 6, data, len);

  return hptools_device_cmd (fd, (unsigned char *)write_6, 6 + len, 0, 0);
}

static SANE_Status
write_string (int fd, const char *format, ...)
{
  va_list args;
  char buf[1024];

  va_start (args, format);
  vsprintf (buf, format, args);
  return write_data (fd, buf, strlen (buf));
}

/* return values: 0 = ok, 1 = got null response, 2 = got malformed response
 *                3 = read-error
 */
static int
inquire (int fd, int request_code, char request_letter,
	 char reply_letter, int *valp)
{
  u_char buf[128], letter, lastchar;
  SANE_Status status;
  int code;
  size_t len;

  *valp = 0;

  sprintf ((char *) buf, "\033*s%d%c", request_code, request_letter);

  DBG(3, "inquire: write <ESC>*s%d%c\n", request_code, request_letter);

  write_string (fd, (char *) buf);
  len = sizeof (buf);
  memset (buf, 0, sizeof (buf));
  status = read_data (fd, buf, &len);
  if (status != SANE_STATUS_GOOD)
    {
      DBG(1, "inquire: read_data failed (%s)\n", sane_strstatus (status));
      return 3;
    }

  if (sscanf ((char *) buf, "\033*s%d%c%d%c", &code, &letter, valp, &lastchar)
      == 4)
  {
    if ((code == request_code) && (letter == reply_letter) && (lastchar == 'V'))
    {
      return 0;
    }
    else
    {
      DBG(1, "inquire: returned string is malformed (%s)\n", buf);
      return 2;
    }
  }
  if (sscanf ((char *) buf, "\033*s%d%c%c", &code, &letter, &lastchar)
      == 3)
  {
    if ((code == request_code) && (letter== reply_letter) && (lastchar == 'N'))
    {
      DBG(5, "inquire: returned null response (%s)\n", buf);
      return (1);
    }
  }
 DBG(1, "inquire: returned string is malformed (%s)\n", buf);
 return 2;
}


/* return values: 0 = ok, 1 = got null response, 2 = got malformed response
 *                3 = read-error
 */
static int
inquire_cmd (int fd, int inq_id, int *supported,
             int *minval, int *maxval, int *cuval)
{int retval;

 *supported = 0;

 retval = inquire (fd, inq_id, 'R', 'p', cuval);
 if (retval != 0) return (retval);

 retval = inquire (fd, inq_id, 'L', 'k', minval);
 if (retval != 0) return (retval);

 retval = inquire (fd, inq_id, 'H', 'g', maxval);
 if (retval != 0) return (retval);

 *supported = 1;

 return 0;
}


/* return values: 0 = ok, 1 = got null response, 2 = got malformed response
 *                3 = read-error
 */
static int
inquire_devparam (int fd, int inq_id, char *response)

{u_char buf[128], null_response[32];
 SANE_Status status;
 size_t len, null_len;
 int i_id, i_val;
 char i_char;

 sprintf ((char *) buf, "\033*s%dE", inq_id);
 sprintf ((char *) null_response, "\033*s%ddN", inq_id);
 null_len = strlen ((char *)null_response);

 DBG (3, "inquire_devparam: write <ESC>%s\n", buf+1);

 write_string (fd, (char *) buf);
 len = sizeof (buf);
 memset (buf, 0, len);
 status = read_data (fd, buf, &len);
 if (status != SANE_STATUS_GOOD)
 {
   DBG(1, "inquire_devparam: read_data failed (%s)\n",
          sane_strstatus (status));
   return 3;
 }

 DBG (3, "inquire_devparam: read <ESC>%s\n", buf+1);

 if (strncmp (buf, null_response, null_len) == 0)
   return 1;

 if (sscanf ((char *)buf, "\033*s%dd%d%c", &i_id, &i_val, &i_char) != 3)
 {
   DBG(1, "inquire_devparam: could not decode response (%s)\n", buf);
   return 2;
 }
 if (i_char == 'V')
   sprintf (response, "%d", i_val);
 else if (i_char == 'W')
 {
   strncpy (response, index (buf, 'W')+1, i_val);
   response[i_val] = '\0';
 }
 else
 {
   strncpy (response, buf, sizeof (buf));
   response[sizeof (buf) - 1] = '\0';
 }

 return 0;
}

static SANE_Status 
do_scan (int fd, char *scan_cmd, int otyp, int dwidth)

{int width, height, nbsl, nread, err = 0;
 char response[512];
 SANE_Status status;
 FILE *out;
 char buf[32*1024];

 if (inquire_devparam (fd, 1024, response) != 0)
 {
   DBG(0, "do_scan: Error reading pixels/scan line\n");
   return SANE_STATUS_INVAL;
 }
 sscanf (response, "%d", &width);

 if (inquire_devparam (fd, 1025, response) != 0)
 {
   DBG(0, "do_scan: Error reading bytes/scan line\n");
   return SANE_STATUS_INVAL;
 }
 sscanf (response, "%d", &nbsl);

 if (inquire_devparam (fd, 1026, response) != 0)
 {
   DBG(0, "do_scan: Error reading number of scan lines\n");
   return SANE_STATUS_INVAL;
 }
 sscanf (response, "%d", &height);
 
 status = write_data (fd, scan_cmd, strlen (scan_cmd));
 if (status != SANE_STATUS_GOOD)
 {
   DBG(0, "do_scan: Error sending scan command\n");
   return SANE_STATUS_INVAL;
 }

 out = fopen (output, "wb");
 if (out == NULL)
 {
   DBG(0, "do_scan: Error open output file %s\n", output);
   return SANE_STATUS_INVAL;
 }

 if (   (otyp == 0) || (otyp == 3)
     || ((otyp == 4) && (dwidth == 8))
     || ((otyp == 5) && (dwidth == 24)))
 {
   int ptyp;

   if ( otyp < 4 )       ptyp = 4;  /* b/w */
   else if ( otyp == 4 ) ptyp = 5;  /* grey */
   else                  ptyp = 6;  /* color */

   fprintf (out, "P%d\n%d %d\n", ptyp, width, height);
   if ( ptyp != 4 ) fprintf (out, "255\n"); /* Write maxval if necessary */

   nread = nbsl * height;
   while (nread > 0)
   {size_t sz;

     sz = sizeof (buf);
     status = read_data (fd, (u_char *)buf, &sz);
     if (status != SANE_STATUS_GOOD)
     {
       DBG(0, "do_scan: Error reading data from scanner\n");
       err = 1;
       break;
     }
     if (sz > 0)
       fwrite (buf, 1, sz, out);
     nread -= sz;
   }
 }
 else
 {unsigned char *scanline;
  int pix, pixval, maxval, ptyp = 3;  /* ASCII color */

   if ( otyp == 4 )
   {
     ptyp = 2;  /* ASCII gray */
     maxval = (1 << dwidth) - 1;
   }
   else
   {
     maxval = (1 << (dwidth/3)) -1;
   }
   fprintf (out, "P%d\n%d %d\n%d\n", ptyp, width, height, maxval);

   scanline = (unsigned char *)malloc (nbsl);
   if (scanline == NULL)
   {
     DBG(0, "do_scan: Not enough memory to store scanline\n");
     return SANE_STATUS_INVAL;
   }
   while (height-- > 0) /* For all scanlines */
   {
     nread = nbsl;
     while (nread > 0)  /* Collect one scanline */
     {size_t sz;

       sz = nread;
       status = read_data (fd, (u_char *)scanline+nbsl-nread, &sz);
       if (status != SANE_STATUS_GOOD)
       {
         DBG(0, "do_scan: Error reading data from scanner\n");
         err = 1;
         break;
       }
       nread -= sz;
     }

     /* Got the scanline. Write ASCII PPM */
     for (pix = 0; pix < nbsl; pix += 2)
     {
       pixval = ((int)scanline[pix])*256 + scanline[pix+1];
       pixval = maxval + 1 - pixval;
       if (pixval < 0) pixval = 0;
       fprintf (out, "%d ", pixval);
     }
   }

   free (scanline);
 }
 
 fclose (out);
 return err ? SANE_STATUS_INVAL : SANE_STATUS_GOOD;
}

static int
do_data (int fd, const char *filename)

{FILE *datafd = fopen (filename, "rb");
 char buf[HP_BUFSIZE];
 int ndata;
 SANE_Status status = SANE_STATUS_INVAL;
 
 if (datafd == NULL) return status;

 ndata = fread (buf, 1, sizeof (buf), datafd);
 fclose (datafd);

 if (ndata > 0)
   status = write_data (fd, buf, ndata);

 return status;
}


static int
do_error (int fd)

{char response[256];
 int err;

  /* Inquire error stack */
  if (inquire_devparam (fd, 257, response) != 0) return -1;
  if (strcmp (response, "0") == 0) return 0;  /* No error */
  /* Inquire current error */
  if (inquire_devparam (fd, 259, response) != 0) return -1;
  if (sscanf (response, "%d", &err) != 1) return -1;
  return err;
}

static int
do_dev_param (int fd, int argc, char **argv)

{int k;
 char response[256];

 if (argc < 0) return -1;
 if ((argv[0][0] != '-') || (argv[0][1] != 'q')) return -1;

 for (k = 0; k < (int)(sizeof (inq_devparam) / sizeof (inq_devparam[0])); k++)
 {
   /* static struct { int inq_id; char *key; char *dsc; } */
   if (inq_devparam[k].key == NULL) continue;
   if (strcmp (inq_devparam[k].key, argv[0]+2) == 0)  /* Inquiry ? */
   {
     if (inquire_devparam (fd, inq_devparam[k].inq_id, response) != 0)
     {
       DBG(0, "do_dev_param: Error during inquiry for %s\n", argv[0]);
       return -1;
     }
     printf ("%s%s", add_str, response);
     add_str = " ";
     return 1;
   }
 }
 return -1;
}


static int
do_scl_command (int fd, int argc, char **argv)

{int k, val;
 SANE_Status status;
 static int dwidth = 24;
 static int otyp = 5;

 if (strcmp (argv[0], "-dwidth") == 0)
   sscanf (argv[1], "%d", &dwidth);

 if (strcmp (argv[0], "-otyp") == 0)
   sscanf (argv[1], "%d", &otyp);

 if (argc < 0) return -1;
 if (argv[0][0] != '-') return -1;

 for (k = 0; k < (int)(sizeof (inq_cmd) / sizeof (inq_cmd[0])); k++)
 {
   /* static struct { char *key; char *dsc; char *code; int inq_id; } */
   if (inq_cmd[k].key == NULL) continue;
   if (strcmp (inq_cmd[k].key, argv[0]+1) == 0)  /* do the command */
   {
     if (argc <= 1)
     {
        DBG(0, "do_scl_command: No value specified\n");
        return -1;
     }
     if (sscanf (argv[1], "%d", &val) != 1)
     {
        DBG(0, "do_scl_command: Bad value specified\n");
        return -1;
     }
     if (   (inq_cmd[k].inq_id == 10484)
         || (inq_cmd[k].inq_id == 10949)
         || (inq_cmd[k].inq_id == 10964))
       status = do_scan (fd, inq_cmd[k].code, otyp, dwidth);
     else
     {char msg[256];

       sprintf (msg, inq_cmd[k].code, val);
       DBG(3, "do_scl_command: write <ESC>%s\n", msg+1);

       status = write_string (fd, inq_cmd[k].code, val);
     }

     if (status != SANE_STATUS_GOOD)
     {
       DBG(0,"do_scl_command: Error executing command\n");
       return -1;
     }
     else
       return 2;
   }
   else if (   (strcmp (inq_cmd[k].key, argv[0]+2) == 0)  /* Inquiry ? */
            && (   (argv[0][1] == 'q') || (argv[0][1] == 'Q')
                || (argv[0][1] == 'C')))
   {int supported, minval, maxval, cuval;

     if (inquire_cmd (fd, inq_cmd[k].inq_id, &supported, &minval, &maxval,
                      &cuval) == 0)
     {
       if (argv[0][1] == 'q') printf ("%s%d", add_str, minval);
       else if (argv[0][1] == 'Q') printf ("%s%d", add_str, maxval);
       else printf ("%s%d", add_str, cuval);
       add_str = " ";
       return 1;
     }
     else
     {
       DBG(0, "do_scl_command: Command not supported by scanner\n");
       return -1;
     }
   }
 }
 return -1;
}


static int
do_command (int fd, int argc, char **argv)

{int nargs = -1;

 /* Try the SCL-commands */
 nargs = do_scl_command (fd, argc, argv);
 if (nargs > 0) return nargs;
 
 nargs = do_dev_param (fd, argc, argv);
 if (nargs > 0) return nargs;

 /* Try other commands */
 if ((strcmp (argv[0], "-output") == 0) && (argc > 1))
 {
   strcpy (output, argv[1]);
   return 2;
 }
 else if (strcmp (argv[0], "-qerror") == 0)
 {
   printf ("%s%d", add_str, do_error (fd));
   add_str = " ";
   return 1;
 }
 else if (strcmp (argv[0], "-data") == 0)
 {
   do_data (fd, argv[1]);
   return 2;
 }

 return -1;
}

static SANE_Status
hp_scan (int argc, char *argv[])
{
  unsigned char result[INQ_LEN];
  char msg[256];
  int err, nargs, k;
  int fd;
  SANE_Status status;
  size_t size;
  HpConnect connect = HP_CONNECT_SCSI;
  int got_connect = 0;

  if (argc >= 2)
  {
    if (strcmp (argv[1], "connect-scsi") == 0)
    { got_connect = 1;
      connect = HP_CONNECT_SCSI;
    }
    else if (strcmp (argv[1], "connect-ptal") == 0)
    { got_connect = 1;
      connect = HP_CONNECT_PTAL;
    }
    else if (strcmp (argv[1], "connect-device") == 0)
    { got_connect = 1;
      connect = HP_CONNECT_DEVICE;
    }
  }

  DBG_INIT ();

  DBG(3, "hp_scan: opening %s\n", argv[0]);
  status = hptools_device_open (argv[0], connect, &fd, 0, 0);
  if (status != SANE_STATUS_GOOD)
    {
      DBG(1, "hp_scan: open failed (%s)\n", sane_strstatus (status));
      return SANE_STATUS_INVAL;
    }

  argc--;
  argv++;
  if (got_connect)
  {
    argc--;
    argv++;

  }

  DBG(3, "hp_scan: sending INQUIRY\n");
  size = sizeof (result);
  status = hptools_device_cmd (fd, inquiry, sizeof (inquiry),
                               (unsigned char *)result,&size);

  msg[0] = '\0';
  for (k = 0; (k < (int)sizeof (result)) && (k < (int)size) && (k < 36); k++)
  {
    if ((result[k] < ' ') || (result[k] > 127))
      sprintf (msg+strlen(msg), "0x%02x", (int)result[k]);
    else
      sprintf (msg+strlen(msg), "%c", result[k]);
  }
  DBG(3, "hp_scan: inquiry returned string (size=%d):\n<%s>\n",
      (int) size, msg);

  if (status != SANE_STATUS_GOOD || size < 16)
    {
      DBG(1, "hp_scan: inquiry failed (%s)\n", sane_strstatus (status));
      hptools_device_close (fd);
      return status;
    }

  result[36] = '\0';

  DBG(3, "hp_scan: sending TEST_UNIT_READY\n");
  status = hptools_device_cmd (fd, test_unit_ready, sizeof (test_unit_ready),
			   0, 0);
  if (status != SANE_STATUS_GOOD)
    {
      DBG(1, "hp_scan: test unit ready failed (%s)\n",
	  sane_strstatus (status));
      hptools_device_close (fd);
      return status;
    }

  err = 0;
  while ((err == 0) && (argc > 0))
  {
    nargs = do_command (fd, argc, argv);
    if (nargs <= 0)
    {
      DBG(0,"hp_scan: error with argument %s\n", argv[0]);
      err = 1;
    }
    else
    {
      argc -= nargs;
      argv += nargs;
    }
  }

  hptools_device_close (fd);

  return err ? SANE_STATUS_INVAL : SANE_STATUS_GOOD;
}


static void usage (int argc, char *argv[])

{int k;

 printf ("Usage: %s <scanner-device> [<connect>] <options>\n", argv[0]);
 printf ("<connect>: connect-scsi|connect-device|connect-ptal\n");
 printf ("<options>: -<key> <value> | -<q|Q|C><key> | -q<devkey> | <addopt>\n");

 printf ("<key> (can be used to set or inquire min/max/current value):\n");
 for (k = 0; k < (int)(sizeof (inq_cmd) / sizeof (inq_cmd[0])); k++)
 {
   if (inq_cmd[k].key == NULL) continue;
   printf ("%-20s(%s)\n", inq_cmd[k].key, inq_cmd[k].dsc);
 }
 printf ("\nUse -q<key> to inquire minimum supported value\n");
 printf ("Use -Q<key> to inquire maximum supported value\n");
 printf ("Use -C<key> to inquire current value\n");

 printf ("\n<devkey> (can be used to inquire device parameters):\n");
 for (k = 0; k < (int)(sizeof (inq_devparam) / sizeof (inq_devparam[0])); k++)
 {
   if (inq_devparam[k].key == NULL) continue;
   printf ("%-20s(%s)\n", inq_devparam[k].key, inq_devparam[k].dsc);
 }
 printf ("\n<addopt>:\n");
 printf ("-output <filename>: set output filename (out.pnm)\n");
 printf ("-qerror: inquire current error number\n");
 printf ("-data <filename>: send data from file to scanner\n");
 printf ("\nOptions are processed in order of their appearance.\n");
}


int main (int argc, char *argv[])

{SANE_Status status;

 if (argc < 3)
 {
   usage (argc, argv);
   exit (1);
 }

 argc--;
 argv++;
 status = hp_scan (argc, argv);

 exit ((status == SANE_STATUS_GOOD) ? 0 : 1);
}
