/* sane - Scanner Access Now Easy.
   Copyright (C) 1997 David Mosberger-Tang/Peter Kirchgessner
   Copyright (C) 2000 David Paschal
   This file is just for testing HP-Scanners for supported commands.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.

   This file contains changes common to hpinq.c and hpscan.c to support
   devices accessed via the PTAL library.

   *NO WARRANTY*   */


#ifndef HPTOOLS_H
#define HPTOOLS_H

#ifdef HAVE_PTAL

#include <ptal.h>

static ptalDevice_t hptools_dev;
static ptalChannel_t hptools_chan;

#endif

/* From ../backend/hp.h */
typedef enum { HP_CONNECT_SCSI, HP_CONNECT_DEVICE,
      HP_CONNECT_PIO, HP_CONNECT_RESERVE,
      HP_CONNECT_PTAL } HpConnect;

static HpConnect hptools_connect = HP_CONNECT_SCSI;

static SANE_Status hptools_device_open (const char *device_name, 
    HpConnect connect, int *fd,
    SANEI_SCSI_Sense_Handler sense_handler,void *sense_arg)
{
   SANE_Status status = SANE_STATUS_INVAL;
   int flags;

#ifdef _O_RDWR
 flags = _O_RDWR;
#else
 flags = O_RDWR;
#endif
#ifdef _O_EXCL
 flags |= _O_EXCL;
#else
 flags |= O_EXCL;
#endif
#ifdef _O_BINARY
 flags |= _O_BINARY;
#endif
#ifdef O_BINARY
 flags |= O_BINARY;
#endif

   hptools_connect = connect;  /* Save connect */

   switch (hptools_connect)
   {
     case HP_CONNECT_PTAL:
#ifdef HAVE_PTAL
       ptalInit();

       hptools_dev=ptalDeviceOpen((char *)device_name);
       if (!hptools_dev) {
         return SANE_STATUS_UNSUPPORTED;
       }
       hptools_connect=HP_CONNECT_PTAL;

       hptools_chan=ptalChannelAllocate(hptools_dev);
       if (!hptools_chan) {
         return SANE_STATUS_NO_MEM;
       }

       ptalChannelSetRemoteService(hptools_chan,PTAL_STYPE_SCAN,0,0);

       if (ptalChannelOpen(hptools_chan)==PTAL_ERROR) {
         return SANE_STATUS_UNSUPPORTED;
       }

       if (ptalChannelPrepareForSelect(hptools_chan,fd,0,0,0,0)==PTAL_ERROR) {
         return SANE_STATUS_UNSUPPORTED;
       }
       status = SANE_STATUS_GOOD;
#else
       DBG(0,"hptools_device_open: not compiled with PTAL-support\n");
       status = SANE_STATUS_UNSUPPORTED;
#endif
       break;

     case HP_CONNECT_SCSI:
       status = sanei_scsi_open(device_name,fd,sense_handler,sense_arg);
       break;

     case HP_CONNECT_DEVICE:
       *fd = open (device_name, flags);
       if (*fd < 0)
       {
         DBG(1, "hptools_device_open: open device %s failed (%s)\n",
             device_name, strerror (errno) );
         status = (errno == EACCES) ?
                   SANE_STATUS_ACCESS_DENIED : SANE_STATUS_INVAL;
       }
       else
       {
         status = SANE_STATUS_GOOD;
       }
       break;

     default:
       DBG(1,"hptools_device_open: unknown connect=%d\n", hptools_connect);
       status = SANE_STATUS_UNSUPPORTED;
      break;
   }
   return status;
}

static void hptools_device_close (int fd)
{

   switch (hptools_connect) {

      case HP_CONNECT_SCSI:
        sanei_scsi_close(fd);
      break;

#ifdef HAVE_PTAL
      case HP_CONNECT_PTAL:
        ptalChannelClose(hptools_chan);
      break;
#endif

      case HP_CONNECT_DEVICE:
        close (fd);
        break;

      default:
        DBG(1,"hptools_sanei_scsi_close: unknown connect=%d\n",
            hptools_connect);
        break;
   }
}

#define HPTOOLS_MAX_REPLY_BUFFER 128

static SANE_Status hptools_device_cmd (int fd,
    const unsigned char * src, size_t src_size,
    unsigned char * dst, size_t * dst_size)

{SANE_Status status = SANE_STATUS_INVAL;
 int len, n, i;

   switch (hptools_connect) {
      case HP_CONNECT_SCSI:
          status = sanei_scsi_cmd(fd,src,src_size,dst,dst_size);
        break;

      case HP_CONNECT_DEVICE:
        switch (src[0])
        {
          case 0x08:  /* read into dest */
            len = (src[2]<<16)+(src[3]<<8)+(src[4]);
            n = read (fd, dst, len);
            if (n == 0) status = SANE_STATUS_EOF;
            else if (n < 0) status = SANE_STATUS_IO_ERROR;
            *dst_size = n;
          break;

          case 0x0a:  /* write from src+6 */
            len = (src[2]<<16)+(src[3]<<8)+(src[4]);
            n = write (fd, src+6, len);
            if (n == 0) status = SANE_STATUS_EOF;
            else if (n < 0) status = SANE_STATUS_IO_ERROR;
            else status = SANE_STATUS_GOOD;
            break;

          case 0x12:  /* Inquiry */
            len = *dst_size;
            for (i = 0; i < len; i++) dst[i]='?';
            status = SANE_STATUS_GOOD;
          break;

          case 0x00:  /* Test unit ready */
            status = SANE_STATUS_GOOD;
          break;
        }
        break;

#ifdef HAVE_PTAL
      case HP_CONNECT_PTAL: {
         struct timeval continueTimeout;
         int r;

           switch (src[0]) {
             case 0x08:   /* read into dst */
             continueTimeout.tv_sec=1;
             continueTimeout.tv_usec=0;
             r=ptalSclChannelRead(hptools_chan,
                dst,*dst_size,0,&continueTimeout,
                *dst_size<=HPTOOLS_MAX_REPLY_BUFFER);
             if (!r) return SANE_STATUS_EOF;
             if (r<0) return SANE_STATUS_IO_ERROR;
             *dst_size=r;
             return SANE_STATUS_GOOD;

             case 0x0A:   /* write from src+6 */
#if 0
             startTimeout.tv_sec=0;
             startTimeout.tv_usec=100000;
             continueTimeout.tv_sec=2;
             continueTimeout.tv_usec=0;
             ptalChannelFlush(hptools_chan,
                &startTimeout,&continueTimeout);
#endif
             len=(src[2]<<16)+(src[3]<<8)+(src[4]);
             r=ptalChannelWrite(hptools_chan,(char *)(src+6),len);
             if (!r) return SANE_STATUS_EOF;
             if (r<0) return SANE_STATUS_IO_ERROR;
             return SANE_STATUS_GOOD;

             case 0x12:   /* INQUIRY */
             r=*dst_size;
             for (i=0;i<r;i++) dst[i]='?';
             return SANE_STATUS_GOOD;

             case 0x00:   /* TEST_UNIT_READY */
             return SANE_STATUS_GOOD;

             default:
             DBG(1,"hptools_device_cmd: "
                "unknown opcode=0x%2.2X\n",src[0]);
             break;
        }
      }
      break;
#endif

      default:
        DBG(1,"hptools_device_cmd: unknown connect=%d\n",
            hptools_connect);
        status = SANE_STATUS_INVAL;
        break;
   }

   return status;
}

#endif   /* HPTOOLS_H */
